/*	===========================================================================
 UPC - EETAC - CSD - EMC - http://digsys.upc.edu

 * LAB5: Digital inputs and outputs
 * FSM architecture, external interrupts, CLK concept, synchronisation
 * Documentation location: https://digsys.upc.edu/csd/DEE/lab05/lab05.html
 * Hardware for prototyping: Arduino Uno board. Chip ATmega328P

Planning: 

- Step 1: Copy and adapt the previous project LAB5 where interrupts and used

- Step 2: Organise the software around a FSM in the same way 
          that similar projects are solved in CSD:
		  https://digsys.upc.edu/csd/P10/lab10/lab10.html#PlanX

=============================================================================*/

/*	===========================================================================
	Include section
	=========================================================================*/

// External function libraries 


/*	===========================================================================
	Definitions
	======================================================================== */
// Switches 
const char pin_CE_L = 5;  
const char pin_Code = 4;


// External CLK for running the counter --> microcontroller's external INT0
const char INT0_pin_CLK_L_BUTTON = 2;  // 

// LEDs for indicating the current state
const char pin_Q2 = 10; 
const char pin_Q1 = 9; 
const char pin_Q0 = 8; 

// LED for indicating terminal count
const char pin_TC8_L = 12;  
 
 
// FSM list of states
const char Num_0	= 'A'; 
const char Num_1	= 'B'; 
const char Num_2	= 'C'; 
const char Num_3	= 'D'; 
const char Num_4	= 'E'; 
const char Num_5	= 'F'; 
const char Num_6	= 'G'; 
const char Num_7	= 'H'; 
const char Num_8	= 'I'; 
const char Num_9	= 'J'; 


/*	===========================================================================
	Global RAM variables
	======================================================================== */

char var_CE;
char var_Code;
char var_TC8;
char var_Q; 
					
// The state variable which will be updated ("asynchronously") every time that 
// "state_logic()" is executed. 
 
char var_current_state;	


//The flag that will be set from the interrupt service routine to indicate that 
// a new CLK_Flag has arrived when detecting a falling edge
static char var_CLK_flag;	

char var_USART_flag; // For writing or sending ASCII messages only one time					


 
/*	===========================================================================
	System initialisation
	=========================================================================*/
void setup() { 
// Configuring digital output pins
	pinMode(pin_TC8_L, OUTPUT);  
	pinMode(pin_Q2, OUTPUT);
	pinMode(pin_Q1, OUTPUT);
	pinMode(pin_Q0, OUTPUT);	


// Configuring digital input pins
	pinMode(pin_CE_L, INPUT);  
	pinMode(pin_Code, INPUT);  

// This is the special pin to be an external interrupt source INT0:
	pinMode(INT0_pin_CLK_L_BUTTON, INPUT_PULLUP); 
	
 //Initialise the FSM and the timer 
	var_current_state = Num_0;	
	
	
// Serial monitor initialisation
	Serial.begin(9600); 
 
// Serial monitor ASCII message:
	Serial.println("Project LAB5- DEE");
	Serial.println("Basic FSM + interrupts: 3-bit counter"); 
	

//	---------------------------------------------------------------------------
// This is the interrupt handler. Read the Arduino documentation at: 
// https://www.arduino.cc/reference/en/language/functions/external-interrupts/attachinterrupt/
// attachInterrupt(digitalPinToInterrupt(pin), ISR, mode) (recommended)
	
	attachInterrupt(digitalPinToInterrupt(INT0_pin_CLK_L_BUTTON), interrupt_service_routine_INT0, FALLING);
	// Interrupt when detecting a falling edge from '1' to '0'
}

/*	===========================================================================
	Infinite loop for solving the FSM
	======================================================================== */
void loop() {       
	read_inputs(); 		// capture the switch voltage values into RAM variables
	if (var_CLK_flag == 1){		
		state_logic();    	// calculating the next state to go
		var_CLK_flag = 0;	// Clear the var_CLK_flag 
	}
	output_logic();		// calculating outputs 
	write_outputs(); // Convert RAM variable values into pin output voltage
}

/*	===========================================================================
	//State variable logic routine  (The CC1)
	======================================================================== */
char state_logic(void){
unsigned char error = 0;	 	
	switch (var_current_state) {
		case Num_0:				
			if (var_CE == 1){
				var_current_state = Num_1;                         
			}
			else{
				var_current_state = Num_0;  
			}                    
		break;

		case Num_1:				
			if (var_CE == 1) var_current_state = Num_2;                         
			else var_current_state = Num_1;                    
		break;
		
		case Num_2:		//else is not required. Check if you see 
						// differences in the disassembly listing 
			if (var_CE == 1) 
				var_current_state = Num_3;                         
		break;

		case Num_3:				
			if (var_CE == 1) 
				var_current_state = Num_4;    
		break;
		case Num_4:				
			if (var_CE == 1) 
				var_current_state = Num_5;    
		break;
		case Num_5:				
			if (var_CE == 1) 
				var_current_state = Num_6;    
		break;
		case Num_6:				
			if (var_CE == 1) 
				var_current_state = Num_7;    
		break;
		case Num_7:				
			if (var_CE == 1) 
				var_current_state = Num_0;    
		break;

		default:
		// If no states have been selected, there is an error
		// and the systems must be stopped                     
			error = 1;	
			break;
	}
	return (error);
}


/*	===========================================================================
//Output logic routine (The combinational circuit CC2)
	======================================================================== */
char output_logic(void){
unsigned char error = 0;	 
if (var_Code == 1) {  // radix-2
	switch (var_current_state) {	
		case Num_0:
            var_Q = 0;
			break;
		case Num_1:
            var_Q = 1;
			break;
		case Num_2:
            var_Q = 2;
			break;
		case Num_3:
            var_Q = 3;
			break;
		case Num_4:
            var_Q = 4;
			break;
		case Num_5:
            var_Q = 5;
			break;
		case Num_6:
            var_Q = 6;
			break;
		case Num_7:
            var_Q = 7;
			break;
		default:
			error = 1;	// If no states have been selected, there is an error
						// and the systems must be stopped 
			break;
	}
}
else {			// Gray

	switch (var_current_state) {	
		case Num_0:
            var_Q = 0;
			break;
		case Num_1:
            var_Q = 1;
			break;
		case Num_2:
            var_Q = 3;
			break;
		case Num_3:
            var_Q = 2;
			break;
		case Num_4:
            var_Q = 6;
			break;
		case Num_5:
            var_Q = 7;
			break;
		case Num_6:
            var_Q = 5;
			break;
		case Num_7:
            var_Q = 4;
			break;
		default:
			error = 1;	// If no states have been selected, there is an error
						// and the systems must be stopped 
			break;
	}

}


// Notice that '&' is a bitwise AND operation, however '&&' is a logic AND: 
	if (var_current_state == Num_7 && var_CE == 1)
		var_TC8 = 1;
	else
		var_TC8 = 0;
 
	return (error);

}


/*	===========================================================================
	ISR external interrupt INT0 pin 
This function is executed when an active edge is detected at pin PD2 (INT0)
The idea is simply set the global variable
	======================================================================== */
void interrupt_service_routine_INT0 (void){

	var_CLK_flag = 1;  

	// LAB experimentation on debouncing techniques: 
	//delay(5); // wait some time 

}


/*	===========================================================================
	read_inputs 
	======================================================================== */
void read_inputs(void){

   if (digitalRead(pin_CE_L) == 1) // Active-low connected switch
		var_CE = 1;
   else
		var_CE = 0;
   
   if (digitalRead(pin_Code) == 1) // Active-low connected switch
		var_Code = 1;
   else
		var_Code = 0;

}

/*	===========================================================================
	write_outputs 
	======================================================================== */
void write_outputs(void){
static char var_buf;  // local variable for saving partial results

	var_buf = var_Q & 0b00000001; //Masking the bit Q0 of interest
	digitalWrite(pin_Q0, var_buf); 

		
	var_buf = (var_Q & 0b00000010)>>1; //Masking the bit Q1 of interest
	digitalWrite(pin_Q1, var_buf); 

	var_buf = (var_Q & 0b00000100)>>2; //Masking the bit Q2 of interest
	digitalWrite(pin_Q2, var_buf); 

	if (var_TC8 == 1)
		digitalWrite(pin_TC8_L, LOW); // Active-low connected LED 
	else
		digitalWrite(pin_TC8_L, HIGH);  

}


